#!/usr/bin/perl -w

# Copyright 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use File::Spec;
use List::Util 'max','min','sum';
use Math::OEIS::Stripped;
use Math::PlanePath::Base::Digits 'bit_split_lowtohigh';
use FLAT;

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyFLAT;
use MyAlternate ':all';
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;

# Richard A. Dean, "A Sequence Without Repeats on x, x-1, y, y-1", American
# Mathematical Monthly, volume 72, number 4, April 1965, pages 383-385.
#
# symbols x, y, x^-1, y^-1
# no 2-block  a^2 a^-e  for a = x or y and e=+/-1
# and no repeat blocks length k
#
#         x, y, x^-1, y^-1
# numbers 1  2  3     4
# quarter blocks ABCD -> ABCD ADCB
#
# A003324
# 1 2 3 4 1 4 3 2 1 2 3 2 1 4 3 4 1 2 3 4 1 4 3 4 1 2 3 2 1 4 3 2 1 2 3 4 1 4 3 2 1 2 3 2 1 4 3 2 1 2 3 4 1 4 3 4 1 2 3 2 1 4 3 4.
#

my @table = ([1,2],
             [3,4],
             [1,4],
             [3,2]);
sub n_by_table {
  my ($n) = @_;
  my $state = 1;
  foreach my $bit (reverse bit_split_lowtohigh($n)) {
    $state = $table[$state-1]->[$bit];
  }
  return $state;
}

# starting n=0,
# bit above lowest 0 and parity of its bit position -> num 2,4
# if even, so lowest 0 at least significant -> num 1,3
sub n_by_bits {
  my ($n) = @_;
  ### n_by_bits(): $n
  if ($n & 1) {
    my $flip = 2;
    do {
      ### at: $n
      $n >>= 1; $flip ^= 2;
    } while ($n & 1);
    ### final: "n=$n flip=$flip"
    return ($flip ^ ($n&2)) + 2;
  } else {
    return ($n&2) + 1;
  }
}
CHECK {
  # n_by_bits(1) == 2 or die;
  # n_by_bits(3) == 4 or die;
}

# starting n=1,
# bit above lowest 1 and parity of its bit position -> num 2,4
# if odd, so lowest 1 at least significant -> num 1,3
sub n1_by_bits {
  my ($n) = @_;
  ### n_by_bits(): $n
  $n >= 1 or die;

  return ($n%4==1 ? 1 : $n%4==3 ? 3 : $n%4==2 ? ($n/2)%4+1 : turn($n)+3); # good
  return 5/2 + (-1)**$n*(1/2 + turn($n));     # good
  return ($n&1 ? 2-turn($n) : 3+turn($n));    # good
  return 3-($n&1) + (-1)**$n *turn($n);       # good
  return ($n&1 ? ($n&2)+1 : 3+turn($n));      # good
  return (($n&1)+1 - turn($n)    -1)%4+1;     # good
  return (2*TurnRpred($n) + ($n&1)  -1)%4+1;  # good

  # good
  my $s = $n;
  my $flip = 0;
  until ($s & 1) {
    ### at: $s
    $s >>= 1; $flip ^= 2;
  }
  ### final: "n=$s flip=$flip"
  return ( ( $flip + ($s&2) + ($n&1))    -1)%4+1;
}
# GP-DEFINE  n1_by_bits(n) = my(k=valuation(n,2)+1); \
# GP-DEFINE    2*bitxor((n+k)%2,bittest(n,k)) + 2-(n%2);
# GP-DEFINE  n1_by_bits(n) = \
# GP-DEFINE    if(n%2, n%4, my(k=valuation(n,2)+1); 2*((k+bittest(n,k))%2)+2);
# for(n=1,20,print1(n1_by_bits(n)))
# my(v=OEIS_samples("A003324")); \
# vector(#v,n, n1_by_bits(n)) == v
# n1_by_bits(12)

{
  my @array = array_by_replicate(8);
  print join('',@array[0..20]),"\n";
  my @samples = Math::OEIS::Stripped->anum_to_values('A003324');
  print join('',@samples),"\n";
  join('',@array[0..$#samples]) eq join('',@samples)  or die;

  my @by_table = map {n_by_table($_)} 0 .. $#array;
  print join('',@by_table[0..20]),"\n";
  join('',@array) eq join('',@by_table)  or die;

  my @by_bits = map {n_by_bits($_)} 0 .. $#array;
  print join('',@by_bits[0..20]),"\n";
  join('',@array) eq join('',@by_bits)  or die;

  my @by_bits1 = map {n1_by_bits($_)} 1 .. @array;
  print join('',@by_bits1[0..20]),"\n";
  join('',@array) eq join('',@by_bits1)  or die;

  my $f = FLAT::DFA->new;
  $f->add_states(4);
  $f->MyFLAT::set_name('high to low');
  $f->set_starting(0);
  $f->set_accepting(3);
  foreach my $from (1 .. 4) {
    foreach my $bit (0,1) {
      my $to = $table[$from-1]->[$bit];
      $f->add_transition($from-1,$to-1, $bit);
    }
  }
  # $f->MyFLAT::view;

  # low to high
  # any..10 = num 3
  # any..00 = num 1
  # num 2
  # .....10 even 1s = num 4
  # .....00 odd 1s = num 4
  # num 4
  # .....00 even 1s = num 4
  # .....10 odd 1s = num 4
  # ......101  = num 4
  $f = $f->MyFLAT::reverse
    ->MyFLAT::set_name('low to high');
  # $f = $f->MyFLAT::minimize(verbose=>1)
  MyFLAT::FLAT_print_tikz($f);
  $f->MyFLAT::view;

  exit 0;
}

sub array_by_replicate {
  my ($k) = @_;
  my @array = (1,2,3,4);
  foreach (1 .. $k) {
    my $q = @array/4;
    push @array, @array[0 .. $q-1];
    push @array, @array[3*$q .. 4*$q-1];
    push @array, @array[2*$q .. 3*$q-1];
    push @array, @array[1*$q .. 2*$q-1];
  }
  return @array;
}
